import typing as t

import attr
import click
import marshmallow
import pyrsistent
from marshmallow import fields


TYPES = {t.__name__: t for t in [int, str, bool, float]}


def get_jsonschema_type_mapping(type_):
    def _jsonschema_type_mapping(self):
        d = {"type": type_}

        if "description" in self.metadata.keys():
            d["description"] = self.metadata["description"]
        else:
            d["description"] = self.metadata["metadata"]["description"]

        return d

    return _jsonschema_type_mapping


class TypeField(marshmallow.fields.Field):
    def __init__(self, *args, **kwargs):
        self.default = kwargs.get("default", marshmallow.missing)
        super().__init__(*args, **kwargs)

    # pylint: disable=redefined-outer-name
    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return TYPES[value]
        except KeyError:
            if self.default == marshmallow.missing:
                raise
            return self.default

    _jsonschema_type_mapping = get_jsonschema_type_mapping("string")


class OptionNameField(marshmallow.fields.Field):
    # pylint: disable=redefined-outer-name
    def _deserialize(self, value, attr, data, **kwargs):
        return [value]

    _jsonschema_type_mapping = get_jsonschema_type_mapping("string")


class ArgumentNameField(marshmallow.fields.Field):
    # pylint: disable=redefined-outer-name
    def _deserialize(self, value, attr, data, **kwargs):
        return [value]

    _jsonschema_type_mapping = get_jsonschema_type_mapping("string")


class AnyField(marshmallow.fields.Field):
    _jsonschema_type_mapping = get_jsonschema_type_mapping("string")


class OptionSchema(marshmallow.Schema):
    """A command line named option for a new command."""

    param_decls = OptionNameField(
        data_key="name",
        metadata={"description": "Name of the option. Usually prefixed with - or --."},
    )
    type = TypeField(
        metadata={"description": f'Name of the type. {", ".join(TYPES)} accepted.'}
    )
    is_flag = fields.Boolean(
        default=False, metadata={"description": "Whether the option is a boolean flag."}
    )
    help = fields.String(
        default=None, metadata={"description": "Documentation for the option."}
    )
    hidden = fields.Boolean(
        default=False,
        metadata={"description": "Whether the option is hidden from help."},
    )
    required = fields.Boolean(
        default=False, metadata={"description": "Whether the option is required."}
    )
    nargs = fields.Integer(
        metadata={"description": "Number of instances expected. Pass -1 for variadic."}
    )
    multiple = fields.Boolean(
        metadata={"description": "Whether multiple values can be passed."}
    )
    default = AnyField(default=None, metadata={"description": "Default value."})
    choices = fields.List(
        fields.String(),
        metadata={"description": "List of allowed string values."},
        default=None,
    )

    @marshmallow.post_load()
    def make_option(self, validated, partial, many):  # pylint: disable=unused-argument
        choices = validated.pop("choices", None)
        if choices:
            validated["type"] = click.Choice(choices)
        return click.Option(**validated)


class ArgumentSchema(marshmallow.Schema):
    """A command-line positional argument for a new command."""

    param_decls = ArgumentNameField(
        data_key="name", metadata={"description": "Name of the argument."}
    )
    type = TypeField(
        default=str,
        metadata={"description": f'Name of the type. {", ".join(TYPES)} accepted.'},
    )
    required = fields.Boolean(
        default=True, metadata={"description": "Whether the argument is required."}
    )
    nargs = fields.Integer(
        default=None,
        metadata={"description": "Number of instances expected. Pass -1 for variadic."},
    )
    choices = fields.List(
        fields.String(),
        metadata={"description": "List of allowed string values."},
        default=None,
    )

    @marshmallow.post_load()
    def make_argument(
        self, validated, partial, many
    ):  # pylint: disable=unused-argument
        choices = validated.pop("choices", None)
        if choices:
            validated["type"] = click.Choice(choices)
        return click.Argument(**validated)


@attr.dataclass(frozen=True)
class RemapParam:
    new: str
    old: str


class RemapParamSchema(marshmallow.Schema):
    """Translation between the name of a base command's parameter and the name of the new command's parameter."""

    new = fields.String(metadata={"description": "New name of the parameter."})
    old = fields.String(metadata={"description": "Old name of the parameter."})

    @marshmallow.post_load()
    def make_remap(self, validated, partial, many):  # pylint: disable=unused-argument
        return RemapParam(**validated)


@attr.dataclass(frozen=True)
class CommandStage:
    command: str
    remap_params: t.List[RemapParam] = attr.ib(converter=pyrsistent.freeze)
    params: t.Dict[str, str] = attr.ib(converter=pyrsistent.freeze)


class CommandStageSchema(marshmallow.Schema):
    """A single stage of a new command pipeline."""

    command = fields.String(metadata={"description": "Name of the base command"})
    remap_params = fields.List(
        fields.Nested(RemapParamSchema),
        missing=list,
        metadata={
            "description": "Provide new names for the parameters, different from the base command parameters' names"
        },
    )
    params = fields.Dict(
        missing=dict,
        metadata={
            "description": "Mapping from new command param name (str) to value (any json type)."
        },
    )

    @marshmallow.post_load()
    def make(self, validated, partial, many):  # pylint: disable=unused-argument
        return CommandStage(**validated)


@attr.dataclass(frozen=True)
class CommandTest:
    invocation: t.List[str] = attr.ib(converter=pyrsistent.freeze)
    input: str
    output: str


class CommandTestSchema(marshmallow.Schema):
    """A test of a new command."""

    invocation = fields.List(
        fields.String(),
        metadata={
            "description": "Command line arguments to mario. (Don't include `mario`.)"
        },
    )
    input = fields.String(
        metadata={"description": "String passed on stdin to the program."}
    )
    output = fields.String(
        metadata={"description": "Expected string output from the program."}
    )

    @marshmallow.post_load()
    def make(self, validated, partial, many):  # pylint: disable=unused-argument
        return CommandTest(**validated)


@attr.dataclass(frozen=True)
class CommandSpec:
    name: str
    short_help: t.Optional[str]
    help: t.Optional[str]
    arguments: t.List[click.Argument] = attr.ib(converter=pyrsistent.freeze)
    options: t.List[click.Option] = attr.ib(converter=pyrsistent.freeze)
    stages: t.List[CommandStage] = attr.ib(converter=pyrsistent.freeze)
    inject_values: t.List[str] = attr.ib(converter=pyrsistent.freeze)
    tests: t.List[CommandTest] = attr.ib(converter=pyrsistent.freeze)
    section: str
    hidden: bool


class CommandSpecSchema(marshmallow.Schema):
    """A new command."""

    name = fields.String(metadata={"description": "Name of the new command."})
    help = fields.String(
        default=None,
        missing=None,
        metadata={
            "description": "Long-form documentation of the command. Will be interpreted as ReStructuredText markup."
        },
    )
    short_help = fields.String(
        default=None,
        missing=None,
        metadata={"description": "Single-line CLI description."},
    )
    arguments = fields.List(
        fields.Nested(ArgumentSchema),
        missing=list,
        metadata={"description": "Arguments accepted by the new command."},
    )
    options = fields.List(
        fields.Nested(OptionSchema),
        missing=list,
        metadata={"description": "Options accepted by the new command."},
    )
    stages = fields.List(
        fields.Nested(CommandStageSchema),
        metadata={
            "description": "List of pipeline command stages that input will go through."
        },
    )
    inject_values = fields.List(
        fields.String(),
        missing=list,
        metadata={
            "description": (
                "CLI parameters to be injected into the local namespace, accessible by the executing commands."
            )
        },
    )
    tests = fields.List(
        fields.Nested(CommandTestSchema),
        missing=list,
        data_key="tests",
        metadata={"description": "List of specifications to test the new command."},
    )
    section = fields.String(
        missing=None,
        metadata={
            "description": "Name of the documentation section in which the new command should appear."
        },
    )
    hidden = fields.Boolean(
        missing=False, metadata={"description": "Hide this command on the help page."}
    )

    @marshmallow.post_load()
    def make(self, validated, partial, many):  # pylint: disable=unused-argument
        return CommandSpec(**validated)



import collections
import os.path
import urllib.parse

import lxml.etree
import pytest
from tests import helpers

from mario import interpret


SYMBOL = "x"


def test_split_pipestring():
    s = 'x ! y + f"{x!r}"'
    sep = "!"
    assert interpret.split_pipestring(s, sep) == ["x", ' y + f"{x!r}"']


@pytest.mark.parametrize(
    "name, expected",
    [
        ("str.upper", {}),
        ("os.path.join", {"os": os, "os.path": os.path}),
        ("map", {}),
        ("collections.Counter", {"collections": collections}),
        ("urllib.parse.urlparse", {"urllib": urllib, "urllib.parse": urllib.parse}),
        ("lxml.etree.parse", {"lxml": lxml, "lxml.etree": lxml.etree}),
    ],
)
def test_get_module(name, expected):
    assert interpret.build_name_to_module(name) == expected


@pytest.mark.parametrize(
    "string, separator, expected",
    [
        ("a", "!", ["a"]),
        ("ab", "!", ["ab"]),
        ("ab!cd", "!", ["ab", "cd"]),
        ("ab!cd!ef", "!", ["ab", "cd", "ef"]),
        ('a"b!c"d!ef', "!", ['a"b!c"d', "ef"]),
        ("a", "\\", ["a"]),
        ("ab", "\\", ["ab"]),
        ("ab\\cd", "\\", ["ab", "cd"]),
        ("ab\\cd\\ef", "\\", ["ab", "cd", "ef"]),
        ('a"b\\c"d\\ef', "\\", ['a"b\\c"d', "ef"]),
        (f'str.upper ! {SYMBOL} + "z"', "!", ["str.upper", f' {SYMBOL} + "z"']),
    ],
)
def test_split_string_on_separator(string, separator, expected):
    result = list(interpret.split_pipestring(string, separator))
    assert result == expected


def test_no_autocall():
    output = helpers.run(["map", "--no-autocall", "1"], input=b"a\nbb\n").decode()
    assert output == "1\n1\n"


def test_autocall():
    output = helpers.run(["map", "len"], input=b"a\nbb\n").decode()
    assert output == "1\n2\n"


def test_autocall_in_eval():
    helpers.run(["eval", "datetime.datetime.now().isoformat()"]).decode()


def test_autocall_requires_symbol():
    output = helpers.run(["map", "pathlib.Path(x).name"], input=b"a\nbb\n").decode()
    assert output == "a\nbb\n"



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # Must come *after* sphinx.ext.napoleon.
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosectionlabel",
    "mario.sphinx_click",
    "mario.sphinx_marshmallow",
    "mario.sphinx_jsonschema",
]
if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

source_suffix = ".rst"
master_doc = "index"
project = "mario"
year = "2019"
author = "mario contributors"
copyright = "{0}, {1}".format(year, author)
version = release = "0.0.152"

pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/python-mario/mario/issues/%s", "#"),
    "pr": ("https://github.com/python-mario/mario/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"


html_theme = "alabaster"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {"**": ["searchbox.html", "globaltoc.html", "sourcelink.html"]}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = True

autoapi_dirs = ["../src/mario"]

html_theme_options = {"navigation_depth": 5}

autosectionlabel_prefix_document = True



import textwrap

import click

import mario.doc


class CommandInSection(click.Command):
    def __init__(self, *args, section=None, **kwargs):
        self.section = section
        super().__init__(*args, **kwargs)


class ReSTCommand(click.Command):
    """Parse help as rst."""

    def format_help_text(self, ctx, formatter):

        if self.help:
            self.help = mario.doc.rst2text(textwrap.dedent(self.help))
            original_wrap_text = click.formatting.wrap_text
            click.formatting.wrap_text = lambda x, *a, **kw: x
            super().format_help_text(ctx, formatter)
            click.formatting.wrap_text = original_wrap_text

        elif self.short_help:
            original_help = self.help
            self.help = self.short_help
            super().format_help_text(ctx, formatter)
            self.help = original_help


class DocumentedCommand(ReSTCommand, CommandInSection):
    pass



import os

import click.testing
import pytest


@pytest.fixture(name="runner")
def _runner():
    return click.testing.CliRunner()


@pytest.fixture(name="tmp_env")
def _tmp_env(tmp_path):
    env = os.environ.copy()
    env["MARIO_CONFIG_DIR"] = tmp_path
    return env



"""Tests for mario addons."""

import subprocess
import sys

import attr
import docshtest
import pytest

import mario.declarative
import mario.plug
import mario.plugins


REGISTRY = mario.plug.make_plugin_commands_registry()
# pylint: disable=no-member
COMMANDS = REGISTRY.commands.values()
TEST_SPECS = [test_spec for command in COMMANDS for test_spec in command.tests]
REQUIRED_FIELDS = ["tests", "help", "short_help"]


def get_param_id(param):
    """Make a verbose id for a test parameter."""
    if attr.has(type(param)):
        return repr(attr.asdict(param))[:35]
    return repr(param)


command_parametrize = pytest.mark.parametrize("command", COMMANDS, ids=get_param_id)


@pytest.mark.parametrize("test_spec", TEST_SPECS, ids=get_param_id)
def test_command_test_spec(test_spec: mario.declarative.CommandTest):
    """The invocation and input generate the expected output."""

    output = subprocess.check_output(
        [sys.executable, "-m", "mario"] + list(test_spec.invocation),
        input=test_spec.input.encode(),
    ).decode()
    assert output == test_spec.output


@command_parametrize
@pytest.mark.parametrize("field_name", REQUIRED_FIELDS, ids=get_param_id)
def test_command_has_required_fields(command, field_name):
    """Test that the command has all required fields."""
    attribute = getattr(command, field_name)
    assert attribute


@command_parametrize
def test_help(command):
    """The command help passes docshtest."""
    lines = command.help.splitlines(keepends=True)
    for line in lines:
        if line.startswith(" " * 2) and not line.startswith(" " * 4):
            raise IndentationError(f"Line starts with 2 spaces and not 4: \n{line}")
    docshtest.shtest_runner(lines, regex_patterns="")


@command_parametrize
def test_invocation_includes_command(command):
    """The test invocation actually includes the tested command."""

    for test in command.tests:
        message = f"The tested command {command.name} is not in the invocation {test.invocation}."
        assert command.name in test.invocation, message



import importlib
import importlib.resources
import importlib.util
import inspect
import sys
import types
import typing as t
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List

import attr
import importlib_metadata
import toml

from mario import config
from mario import declarative


@attr.dataclass
class PluginObject:
    name: str
    traversal_function: types.FunctionType
    required_parameters: List[str]
    calculate_more_params: types.FunctionType = attr.ib(default=lambda x: {})


@attr.dataclass
class CommandStage:
    name: str

    options: List[str]
    arguments: List[str]
    remap_params: Dict


@attr.dataclass
class CommandCommand:
    name: str
    components: List[CommandStage]
    short_help: str
    options: t.Dict = attr.ib(factory=dict)
    arguments: t.Dict = attr.ib(factory=dict)


@attr.s(repr=False)
class _NoDefaultType:
    def __repr__(self):
        return "NO_DEFAULT"


NO_DEFAULT = _NoDefaultType()


@attr.dataclass
class GlobalOption:
    name: str
    type: t.Type
    default: _NoDefaultType


@attr.s
class Registry:
    traversals: Dict[str, PluginObject] = attr.ib(factory=dict)
    global_options: Dict[str, GlobalOption] = attr.ib(factory=dict)
    cli_functions: Dict[str, Any] = attr.ib(factory=dict)
    commands: Dict[str, CommandCommand] = attr.ib(factory=dict)

    def add_traversal(self, name=None, calculate_more_params=lambda x: {}):
        def wrap(function):

            if name is None:
                registered_name = function.__name__
            else:
                registered_name = name

            params = [param for param in inspect.signature(function).parameters.keys()]
            # pylint: disable=unsupported-assignment-operation
            self.traversals[
                registered_name
            ] = PluginObject(  # pylint: disable=unsupported-assignment-operation
                registered_name, function, params, calculate_more_params
            )
            return function

        return wrap

    def add_cli(self, name=None):
        def wrap(function):
            if name is None:
                registered_name = getattr(function, "__name__", None), function.name
            else:
                registered_name = name
            # pylint: disable=unsupported-assignment-operation
            self.cli_functions[registered_name] = function

            return function

        return wrap


def plugin_module_paths() -> List[str]:
    return [
        entry_point.value + "." + entry_point.name
        for entry_point in importlib_metadata.entry_points()["mario_plugins"]
    ]


def collect_modules(import_paths: Iterable[str]) -> List[types.ModuleType]:
    modules = []
    for path in import_paths:
        modules.append(importlib.import_module(path))
    return modules


def combine_registries(registries):
    global_options = {}
    traversals = {}
    cli_functions = {}
    commands = {}
    for registry in registries:
        traversals.update(registry.traversals)
        global_options.update(registry.global_options)
        cli_functions.update(registry.cli_functions)
        commands.update(registry.commands)
    return Registry(traversals, global_options, cli_functions, commands)


def make_plugin_registry():
    plugin_modules = collect_modules(plugin_module_paths())
    plugin_registries = [module.registry for module in plugin_modules]

    return combine_registries(plugin_registries)


def make_config_registry():
    sys.path.append(str(config.get_config_dir()))

    try:
        # pylint: disable=import-error
        import m
    except ModuleNotFoundError:
        return Registry()

    return getattr(m, "registry", None) or Registry()


def make_global_registry():
    return combine_registries(
        [
            make_plugin_registry(),
            make_config_registry(),
            make_config_commands_registry(),
            make_plugin_commands_registry(),
        ]
    )


def make_commands(conf):

    commands = declarative.CommandSpecSchema(many=True).load(conf.get("command", []))

    return commands


def make_config_commands_registry():
    conf = config.load_config()

    commands = make_commands(conf)
    return Registry(commands={c.name: c for c in commands})


def make_plugin_commands_registry(package="mario.plugins"):
    plugin_tomls = [
        filename
        for filename in importlib.resources.contents(package)
        if filename.endswith(".toml")
    ]
    confs = [
        toml.loads(importlib.resources.read_text(package, filename))
        for filename in plugin_tomls
    ]

    conf_command_groups = [make_commands(conf) for conf in confs]
    registries = [
        Registry(commands={c.name: c for c in commands})
        for commands in conf_command_groups
    ]
    return combine_registries(registries)



import itertools
import textwrap
import traceback

import click
from docutils import nodes
from docutils import statemachine
from docutils.parsers import rst
from docutils.parsers.rst import directives
from sphinx.util import logging

from mario import doc


LOG = logging.getLogger(__name__)
CLICK_VERSION = tuple(int(x) for x in click.__version__.split("."))


def _indent(text, level=1):
    prefix = " " * (4 * level)

    def prefixed_lines():
        for line in text.splitlines(True):
            yield prefix + line if line.strip() else line

    return "".join(prefixed_lines())


def _get_usage(ctx):
    """Alternative, non-prefixed version of 'get_usage'."""
    formatter = ctx.make_formatter()
    pieces = ctx.command.collect_usage_pieces(ctx)
    formatter.write_usage(ctx.command_path, " ".join(pieces), prefix="")
    return formatter.getvalue().rstrip("\n")


def _get_help_record(opt):
    """Re-implementation of click.Opt.get_help_record.

    The variant of 'get_help_record' found in Click makes uses of slashes to
    separate multiple opts, and formats option arguments using upper case. This
    is not compatible with Sphinx's 'option' directive, which expects
    comma-separated opts and option arguments surrounded by angle brackets [1].

    [1] http://www.sphinx-doc.org/en/stable/domains.html#directive-option
    """

    def _write_opts(opts):
        rv, _ = click.formatting.join_options(opts)
        if not opt.is_flag and not opt.count:
            rv += " <{}>".format(opt.name)
        return rv

    rv = [_write_opts(opt.opts)]
    if opt.secondary_opts:
        rv.append(_write_opts(opt.secondary_opts))

    # pylint: disable=redefined-builtin
    help = opt.help or ""
    extra = []
    if opt.default is not None and opt.show_default:
        extra.append(
            "default: %s"
            % (
                ", ".join("%s" % d for d in opt.default)
                if isinstance(opt.default, (list, tuple))
                else opt.default,
            )
        )
    if opt.required:
        extra.append("required")
    if extra:
        help = "%s[%s]" % (help and help + "  " or "", "; ".join(extra))
    if isinstance(opt.type, click.Choice):
        help = "%s\n\n:options: %s" % (
            help and help + "  " or "",
            "|".join(opt.type.choices),
        )

    return ", ".join(rv), help


def _format_description(ctx):
    """Format the description for a given `click.Command`.

    We parse this as reStructuredText, allowing users to embed rich
    information in their help messages if they so choose.
    """
    help_string = ctx.command.help or ctx.command.short_help
    if not help_string:
        return

    help_string = textwrap.dedent(help_string)
    bar_enabled = False
    for line in statemachine.string2lines(
        help_string, tab_width=4, convert_whitespace=True
    ):
        if line == "\b":
            bar_enabled = True
            continue
        if line == "":
            bar_enabled = False
        line = "| " + line if bar_enabled else line
        yield line
    yield ""


def _format_usage(ctx):
    """Format the usage for a `click.Command`."""
    yield ".. code-block:: shell"
    yield ""
    for line in _get_usage(ctx).splitlines():
        yield _indent(line)
    yield ""


def _format_option(opt):
    """Format the output for a `click.Option`."""
    opt = _get_help_record(opt)

    yield ".. option:: {}".format(opt[0])
    if opt[1]:
        yield ""
        for line in statemachine.string2lines(
            opt[1], tab_width=4, convert_whitespace=True
        ):
            yield _indent(line)


def _format_options(ctx):
    """Format all `click.Option` for a `click.Command`."""
    # the hidden attribute is part of click 7.x only hence use of getattr
    params = [
        x
        for x in ctx.command.params
        if isinstance(x, click.Option) and not getattr(x, "hidden", False)
    ]

    for param in params:
        for line in _format_option(param):
            yield line
        yield ""


def _format_argument(arg):
    """Format the output of a `click.Argument`."""
    yield ".. option:: {}".format(arg.human_readable_name)
    yield ""
    yield _indent(
        "{} argument{}".format(
            "Required" if arg.required else "Optional", "(s)" if arg.nargs != 1 else ""
        )
    )


def _format_arguments(ctx):
    """Format all `click.Argument` for a `click.Command`."""
    params = [x for x in ctx.command.params if isinstance(x, click.Argument)]

    for param in params:
        for line in _format_argument(param):
            yield line
        yield ""


def _format_envvar(param):
    """Format the envvars of a `click.Option` or `click.Argument`."""
    yield ".. envvar:: {}".format(param.envvar)
    yield "   :noindex:"
    yield ""
    if isinstance(param, click.Argument):
        param_ref = param.human_readable_name
    else:
        # if a user has defined an opt with multiple "aliases", always use the
        # first. For example, if '--foo' or '-f' are possible, use '--foo'.
        param_ref = param.opts[0]

    yield _indent("Provide a default for :option:`{}`".format(param_ref))


def _format_envvars(ctx):
    """Format all envvars for a `click.Command`."""
    params = [x for x in ctx.command.params if getattr(x, "envvar")]

    for param in params:
        yield ".. _{command_name}-{param_name}-{envvar}:".format(
            command_name=ctx.command_path.replace(" ", "-"),
            param_name=param.name,
            envvar=param.envvar,
        )
        yield ""
        for line in _format_envvar(param):
            yield line
        yield ""


def _format_subcommand(command):
    """Format a sub-command of a `click.Command` or `click.Group`."""

    yield ".. object:: {}".format(command.name)

    # click 7.0 stopped setting short_help by default
    if CLICK_VERSION < (7, 0):
        short_help = command.short_help
    else:
        short_help = command.get_short_help_str()

    if short_help:
        yield ""
        for line in statemachine.string2lines(
            short_help, tab_width=4, convert_whitespace=True
        ):
            yield _indent(line)


def _get_lazyload_commands(multicommand):
    commands = {}
    for command in multicommand.list_commands(multicommand):
        commands[command] = multicommand.get_command(multicommand, command)

    return commands


def _filter_commands(ctx, commands=None):
    """Return list of used commands."""
    lookup = getattr(ctx.command, "commands", {})
    if not lookup and isinstance(ctx.command, click.MultiCommand):
        lookup = _get_lazyload_commands(ctx.command)

    if commands is None:
        return sorted(lookup.values(), key=lambda item: item.name)

    names = [name.strip() for name in commands.split(",")]
    return [lookup[name] for name in names if name in lookup]


# pylint: disable=too-many-branches
def _format_command(ctx, show_nested, commands=None):
    """Format the output of `click.Command`."""
    # the hidden attribute is part of click 7.x only hence use of getattr
    if getattr(ctx.command, "hidden", False):
        return

    # description

    for line in _format_description(ctx):
        yield line

    yield ".. program:: {}".format(ctx.command_path)

    # usage

    for line in _format_usage(ctx):
        yield line

    # options

    lines = list(_format_options(ctx))
    if lines:
        # we use rubric to provide some separation without exploding the table
        # of contents
        yield ".. rubric:: Options"
        yield ""

    for line in lines:
        yield line

    # arguments

    lines = list(_format_arguments(ctx))
    if lines:
        yield ".. rubric:: Arguments"
        yield ""

    for line in lines:
        yield line

    # environment variables

    lines = list(_format_envvars(ctx))
    if lines:
        yield ".. rubric:: Environment variables"
        yield ""

    for line in lines:
        yield line

    # if we're nesting commands, we need to do this slightly differently
    if show_nested:
        return

    commands = _filter_commands(ctx, commands)

    if commands:
        yield ".. rubric:: Commands"
        yield ""

    for command in commands:
        # Don't show hidden subcommands
        if CLICK_VERSION >= (7, 0):
            if command.hidden:
                continue

        for line in _format_subcommand(command):
            yield line
        yield ""


class ClickDirective(rst.Directive):

    has_content = False
    required_arguments = 1
    option_spec = {
        "prog": directives.unchanged_required,
        "show-nested": directives.flag,
        "commands": directives.unchanged,
    }

    def _load_module(self, module_path):
        """Load the module."""
        # __import__ will fail on unicode,
        # so we ensure module path is a string here.
        module_path = str(module_path)

        try:
            module_name, attr_name = module_path.split(":", 1)
        except ValueError:  # noqa
            raise self.error(
                '"{}" is not of format "module:parser"'.format(module_path)
            )

        try:
            mod = __import__(module_name, globals(), locals(), [attr_name])
        except (Exception, SystemExit) as exc:  # noqa
            err_msg = 'Failed to import "{}" from "{}". '.format(attr_name, module_name)
            if isinstance(exc, SystemExit):
                err_msg += "The module appeared to call sys.exit()"
            else:
                err_msg += "The following exception was raised:\n{}".format(
                    traceback.format_exc()
                )

            raise self.error(err_msg)

        if not hasattr(mod, attr_name):
            raise self.error(
                'Module "{}" has no attribute "{}"'.format(module_name, attr_name)
            )

        parser = getattr(mod, attr_name)

        if not isinstance(parser, click.BaseCommand):
            raise self.error(
                '"{}" of type "{}" is not derived from '
                '"click.BaseCommand"'.format(type(parser), module_path)
            )
        return parser

    def _make_subcommand_to_section(self, command):
        subcommand_to_section = {}
        for help_section in command.sections:
            for subcommand_name in help_section.entries:
                subcommand_to_section[subcommand_name] = help_section
        return subcommand_to_section

    def _get_section_spec(self, cmd):

        if cmd.section in doc.SECTION_NAME_TO_SECTION_SPEC:
            return doc.SECTION_NAME_TO_SECTION_SPEC[cmd.section]
        if cmd.section:
            return doc.HelpSectionSpec(
                priority=doc.DEFAULT_SECTION_PRIORITY, name=cmd.section
            )
        return doc.NULL_SECTION

    def _sort_commands(self, command, subcommands):
        if not hasattr(command, "sections"):
            return subcommands

        return sorted(subcommands, key=self._get_section_spec)

    def _group_commands(self, command, subcommands):
        if not hasattr(command, "sections"):
            return subcommands

        return itertools.groupby(subcommands, key=self._get_section_spec)

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-arguments
    def _generate_nodes(
        self, name, command, parent=None, show_nested=False, commands=None
    ):
        """Generate the relevant Sphinx nodes.

        Format a `click.Group` or `click.Command`.

        :param name: Name of command, as used on the command line
        :param command: Instance of `click.Group` or `click.Command`
        :param parent: Instance of `click.Context`, or None
        :param show_nested: Whether subcommands should be included in output
        :param commands: Display only listed commands or skip the section if
            empty
        :returns: A list of nested docutil nodes
        """
        ctx = click.Context(command, info_name=name, parent=parent)

        if CLICK_VERSION >= (7, 0) and command.hidden:
            return []

        # Title

        item = nodes.section(
            "",
            nodes.title(text=name),
            ids=[nodes.make_id(ctx.command_path)],
            names=[nodes.fully_normalize_name(ctx.command_path)],
        )

        # Summary

        source_name = ctx.command_path
        result = statemachine.ViewList()

        lines = _format_command(ctx, show_nested, commands)

        for line in lines:
            LOG.debug(line)
            result.append(line, source_name)

        self.state.nested_parse(result, 0, item)

        # Subcommands

        if not show_nested:
            return [item]

        commands = _filter_commands(ctx, commands)
        commands = self._sort_commands(command, commands)

        for help_section, subcommands in self._group_commands(command, commands):
            group_name = help_section.name

            if group_name == doc.UNSECTIONED:

                for subcommand in subcommands:
                    item.extend(
                        self._generate_nodes(
                            subcommand.name, subcommand, ctx, show_nested
                        )
                    )

                self.state.nested_parse(result, 0, item)
                continue

            group_item = nodes.section(
                "",
                nodes.title(text=group_name),
                ids=[nodes.make_id(group_name)],
                names=[nodes.fully_normalize_name(group_name)],
            )

            group_list = statemachine.ViewList()

            # pylint: disable=fixme
            # XXX This is supposed to add documentation lines to each group, but it doesn't seem to work.
            for line in help_section.doc.splitlines():
                group_list.append(line, group_name)

            for subcommand in subcommands:
                group_item.extend(
                    self._generate_nodes(subcommand.name, subcommand, ctx, show_nested)
                )

            self.state.nested_parse(group_list, 0, group_item)

            item += group_item

        return [item]

    def run(self):
        # pylint: disable=attribute-defined-outside-init
        self.env = self.state.document.settings.env

        command = self._load_module(self.arguments[0])

        if "prog" not in self.options:
            raise self.error(":prog: must be specified")

        prog_name = self.options.get("prog")
        show_nested = "show-nested" in self.options
        commands = self.options.get("commands")

        return self._generate_nodes(prog_name, command, None, show_nested, commands)


def setup(app):
    app.add_directive("click", ClickDirective)



import os
import textwrap

from mario import utils

from . import helpers


def test_base_exec_before(tmp_path):
    config_body = """
    base_exec_before = 's = "value "'
    """
    config_file_path = tmp_path / "config.toml"

    config_file_path.write_text(config_body)

    args = ["--exec-before", "t = 'is '", "map", "s+t+x"]
    stdin = "ab\ncd\n".encode()
    env = dict(os.environ)
    env.update({f"{utils.NAME}_CONFIG_DIR".upper().encode(): str(tmp_path).encode()})
    output = helpers.run(args, input=stdin, env=env).decode()
    assert output == """value is ab\nvalue is cd\n"""


def test_config_jsonl_command(tmp_path):

    config_body = (helpers.TESTS_DIR / "data/config/jsonl_command.toml").read_text()
    config_file_path = tmp_path / "config.toml"

    config_file_path.write_text(config_body)

    args = ["jsonl"]
    stdin = textwrap.dedent(
        """\
    {"a": 1, "b": 2}
    {"a": 3, "b": 4}
    """
    ).encode()
    env = dict(os.environ)
    env.update({f"{utils.NAME}_CONFIG_DIR".upper().encode(): str(tmp_path).encode()})
    output = helpers.run(args, input=stdin, env=env).decode()
    assert output == textwrap.dedent(
        """\
    {'a': 1, 'b': 2}
    {'a': 3, 'b': 4}
    """
    )


def test_m_namespace(tmp_path, tmp_env):
    """The init file is available under the ``m`` namespace."""
    file = tmp_path / "m" / "__init__.py"
    file.parent.mkdir()
    file.write_text("var = 1\n")
    result = helpers.run(["eval", "m.var"], env=tmp_env).decode()
    assert result == "1\n"


def test_m_init_executed_at_startup(tmp_path, tmp_env):
    """The init file is executed at startup time."""
    file = tmp_path / "m" / "__init__.py"
    file.parent.mkdir()
    file.write_text("print('hello')")
    result = helpers.run(["eval", '""'], env=tmp_env).decode()
    assert result == "hello\n\n"


def test_m_submodule(tmp_path, tmp_env):
    """The submodule is available under the ``m`` namespace."""
    file = tmp_path / "m" / "example.py"
    file.parent.mkdir()
    file.write_text("var = 1\n")
    result = helpers.run(["eval", "m.example.var"], env=tmp_env).decode()
    assert result == "1\n"


def test_m_submodule_not_executed_at_startup(tmp_path, tmp_env):
    """The submodule is *not* executed at startup."""
    file = tmp_path / "m" / "example.py"
    file.parent.mkdir()
    file.write_text("print('hello')")
    result = helpers.run(["eval", "''"], env=tmp_env).decode()
    assert result == "\n"


def test_hidden_false(tmp_path, tmp_env):
    """Non-hidden commands do appear in --help."""

    text = """
    [[command]]
    name = "my-visible-command"
    hidden = false

    [[command.stages]]
    command = "eval"
    params = {code='1'}
    """

    (tmp_path / "config.toml").write_text(text)

    result = helpers.run(["--help"], env=tmp_env).decode()
    assert "my-visible-command" in result


def test_hidden_true(tmp_path, tmp_env):
    """Hidden commands don't appear in --help."""

    text = """
    [[command]]
    name = "my-hidden-command"
    hidden = true

    [[command.stages]]
    command = "eval"
    params = {code='1'}
    """

    (tmp_path / "config.toml").write_text(text)

    result = helpers.run(["--help"], env=tmp_env).decode()
    assert "my-hidden-command" not in result



"""Functions for read commands."""

import csv
import typing as t


def read_csv_dicts(file, **kwargs) -> t.Iterable[t.Mapping[t.Any, str]]:
    """Read csv rows into an iterable of dicts."""
    rows = list(file)

    first_row = next(csv.reader(rows))

    fieldnames = first_row
    reader = csv.DictReader(rows, fieldnames=fieldnames, **kwargs)
    return list(reader)[1:]


def read_csv_tuples(file, **kwargs) -> t.Iterable[t.Tuple]:
    """Read csv rows into an iterable of tuples."""
    return (tuple(row) for row in csv.reader(file, **kwargs))



import asyncio
import contextlib
import contextvars
import datetime
import itertools
import json
import sys

import attr
from aiohttp import web


START_TIME = None
counter = itertools.count()
ID: contextvars.ContextVar[int] = contextvars.ContextVar("id")


@attr.s
class Handler:
    async def handle(self, request):
        # pylint: disable=global-statement
        global START_TIME
        if START_TIME is None:
            START_TIME = datetime.datetime.utcnow()

        ID.set(next(counter))
        delay = request.rel_url.query["delay"]

        elapsed = (datetime.datetime.utcnow() - START_TIME).seconds
        print(
            json.dumps(
                dict(message="receive", id=ID.get(), elapsed=elapsed, delay=delay)
            ),
            file=sys.stderr,
        )
        await asyncio.sleep(int(delay))
        elapsed = (datetime.datetime.utcnow() - START_TIME).seconds
        response = json.dumps(
            dict(message="respond", id=ID.get(), elapsed=elapsed, delay=delay)
        )
        print(response, file=sys.stderr)

        return web.Response(text=response)


if __name__ == "__main__":
    with contextlib.redirect_stdout(new_target=sys.stderr):
        app = web.Application()
        handler = Handler()
        handle = handler.handle
        app.add_routes([web.get("/", handle), web.get("/delay/{delay}", handle)])

        web.run_app(app)



from __future__ import annotations

import types
import typing as t

import attr


def parse_rst(text: str) -> docutils.nodes.document:  # type: ignore
    import docutils.nodes
    import docutils.parsers.rst
    import docutils.utils

    parser = docutils.parsers.rst.Parser()
    components = (docutils.parsers.rst.Parser,)
    settings = docutils.frontend.OptionParser(
        components=components
    ).get_default_values()
    document = docutils.utils.new_document("<rst-doc>", settings=settings)
    parser.parse(text, document)
    return document


# pylint: disable=undefined-variable
def format_text(document: docutils.nodes.document) -> str:  # type: ignore
    import sphinx.builders.text
    import sphinx.events
    import sphinx.util.osutil
    import sphinx.writers.text

    app = types.SimpleNamespace(
        srcdir=None,
        confdir=None,
        outdir=None,
        doctreedir="/",
        config=types.SimpleNamespace(
            text_newlines="native",
            text_sectionchars="=",
            text_add_secnumbers=False,
            text_secnumber_suffix=".",
        ),
        tags=set(),
        events=sphinx.events.EventManager(),
        registry=types.SimpleNamespace(
            create_translator=lambda self, something, new_builder: sphinx.writers.text.TextTranslator(
                document, new_builder
            )
        ),
    )

    builder = sphinx.builders.text.TextBuilder(app)

    translator = sphinx.writers.text.TextTranslator(document, builder)
    document.walkabout(translator)
    return translator.body


def rst2text(source: str) -> str:
    document = parse_rst(source)
    return format_text(document)


@attr.dataclass(frozen=True)
class HelpSection:
    priority: int
    entries: t.List[str]
    doc: str = ""
    name: t.Optional[str] = None


@attr.dataclass(frozen=True)
class HelpSectionSpec:
    priority: int
    doc: str = ""
    name: str = ""


UNSECTIONED = "UNSECTIONED"
UNSECTIONED_PRIORITY = 100_000

SECTION_SPECS = {
    HelpSectionSpec(
        name="Traversals", doc="Commands for calling code on data.", priority=0
    ),
    HelpSectionSpec(
        name="Async traversals",
        doc="Commands for asynchronously calling code on data.",
        priority=1,
    ),
    HelpSectionSpec(name=UNSECTIONED, doc="", priority=UNSECTIONED_PRIORITY),
}

SECTION_NAME_TO_SECTION_SPEC = {s.name: s for s in SECTION_SPECS}

DEFAULT_SECTION_PRIORITY = 500

NULL_SECTION = HelpSectionSpec(
    priority=1000, name="Custom", doc="Custom defined commands"
)



import importlib


sphinx_jsonschema = importlib.import_module("sphinx-jsonschema")

_glob_app = None


class DefinitionWideFormat(sphinx_jsonschema.wide_format.WideFormat):  # type: ignore
    """Add definition field."""

    def _objecttype(self, schema):
        # create description and type rows
        rows = self._simpletype(schema)
        rows.extend(self._objectproperties(schema, "definition"))
        rows.extend(self._objectproperties(schema, "properties"))
        rows.extend(self._objectproperties(schema, "patternProperties"))
        rows.extend(self._bool_or_object(schema, "additionalProperties"))
        rows.extend(self._kvpairs(schema, self.KV_OBJECT))
        return rows

    def _dispatch(self, schema, label=None):
        # Main driver of the recursive schema traversal.
        rows = []
        # pylint: disable=no-member
        self.nesting += 1

        if "type" in schema:
            # select processor for type
            if "object" in schema["type"]:
                rows = self._objecttype(schema)
            elif "array" in schema["type"]:
                rows = self._arraytype(schema)
            else:
                rows = self._simpletype(schema)
        else:
            if "description" in schema:
                rows.append(self._line(self._cell(schema["description"])))

        if "$ref" in schema:
            rows.append(
                self._line(
                    self._cell(
                        (":ref:`" + schema["$ref"] + "`").replace("#/definitions/", "")
                    )
                )
            )

        for k in self.COMBINATORS:
            # combinators belong at this level as alternative to type
            if k in schema:
                items = []
                for s in schema[k]:
                    items.extend(self._dispatch(s, self._cell("-")))
                rows.extend(self._prepend(self._cell(k), items))

        for k in self.SINGLEOBJECTS:
            # combinators belong at this level as alternative to type
            if k in schema:
                rows.extend(self._dispatch(schema[k], self._cell(k)))

        # definitions aren't really type equiv's but still best place for them
        rows.extend(self._objectproperties(schema, "definitions"))

        if label is not None:
            # prepend label column if required
            rows = self._prepend(label, rows)

        # pylint: disable=no-member
        self.nesting -= 1
        return rows


class DefinitionJsonSchema(sphinx_jsonschema.JsonSchema):  # type: ignore
    def run(self):
        # pylint: disable=redefined-builtin
        # pylint: disable=protected-access
        format = DefinitionWideFormat(
            self.state, self.lineno, sphinx_jsonschema._glob_app
        )
        return format.transform(self.schema)


def setup(app):

    global _glob_app  # pylint: disable=global-statement
    _glob_app = app
    app.add_directive("jsonschema", DefinitionJsonSchema)



#!/usr/bin/env python
"""Command line pipes in python."""

from __future__ import annotations
from __future__ import generator_stop

import ast
import enum
import importlib
import re
import sys
import textwrap
import types

import attr
import parso


SYMBOL = "x"


class HowCall(enum.Enum):
    NONE = ""
    SINGLE = f"({SYMBOL})"
    VARARGS = f"(*{SYMBOL})"
    VARKWARGS = f"(**{SYMBOL})"


class HowSig(enum.Enum):
    NONE = f"{SYMBOL}"
    SINGLE = f"{SYMBOL}"
    VARARGS = f"*{SYMBOL}"
    VARKWARGS = f"**{SYMBOL}"


howcall_to_howsig = {v: HowSig.__members__[k] for k, v in HowCall.__members__.items()}


@attr.dataclass
class Function:
    wrapped: types.FunctionType
    global_namespace: dict
    source: str

    def __call__(self, *x):
        return self.wrapped(*x)


def _get_named_module(name):
    builtins = sys.modules["builtins"]
    if hasattr(builtins, name):
        return builtins
    try:
        return importlib.import_module(name)
    except ImportError:
        pass
    raise LookupError(f"Could not find {name}")


def _get_autoimport_module(fullname):
    name_parts = fullname.split(".")
    try_names = []
    for idx in range(len(name_parts)):
        try_names.insert(0, ".".join(name_parts[: idx + 1]))

    name_to_module = {}
    for name in try_names:
        try:
            module = _get_named_module(name)
        except LookupError:
            pass
        else:
            if module is sys.modules["builtins"]:
                return {}
            name_to_module[name] = module

    return name_to_module


def find_maybe_module_names(text):
    # pylint: disable=fixme
    # TODO: Use a real parser.
    return re.findall(r"\b[^\d\W]\w*(?:\.[^\d\W]\w*)+\b", text)


def split_pipestring(s, sep="!"):
    segments = []
    tree = parso.parse(s)
    current_nodes = []

    for c in tree.children:
        if isinstance(c, parso.python.tree.PythonErrorLeaf) and c.value == sep:
            segments.append(current_nodes)
            current_nodes = []
        else:
            current_nodes.append(c)

    segments.append(current_nodes)

    return ["".join(node.get_code() for node in seg) for seg in segments]


def make_autocall(expression, howcall):

    # Don't autocall if SYMBOL appears in the expression.
    tree = ast.parse(expression)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name) and node.id == SYMBOL:
            return expression

    return expression + howcall.value


def build_source(components, howcall):
    components = [c.strip() for c in components]
    components = [make_autocall(c, howcall) for c in components]
    indent = "        "
    lines = "".join([f"{indent}{SYMBOL} = {c}\n" for c in components])

    howsig = howcall_to_howsig[howcall]
    source = textwrap.dedent(
        f"""\
    async def _mario_runner({howsig.value}):
{lines}
        return {SYMBOL}
    """
    )

    return source


def build_name_to_module(code):
    name_to_module = {}
    components = split_pipestring(code)
    module_names = {name for c in components for name in find_maybe_module_names(c)}
    for name in module_names:
        name_to_module.update(_get_autoimport_module(name))

    return name_to_module


def build_function(code, global_namespace, howcall):
    name_to_module = build_name_to_module(code)
    global_namespace = {**name_to_module, **global_namespace}

    source = build_source(split_pipestring(code), howcall)
    # pylint: disable=exec-used
    exec(source, global_namespace)
    function = global_namespace["_mario_runner"]
    return Function(function, global_namespace, source)


def build_global_namespace(source):
    if source is None:
        return {}
    global_namespace = {}

    # pylint: disable=exec-used
    exec(source, global_namespace)

    return global_namespace



import os
import pathlib

import appdirs
import toml

from . import interpret
from . import utils


DEFAULTS = {
    "max_concurrent": 5,
    "exec_before": None,
    "autocall": interpret.HowCall.SINGLE,
    "base_exec_before": None,
    "dir_path": os.environ.get(f"{utils.NAME}_CONFIG_DIR".upper(), None),
}


def get_config_dir():
    str_path = DEFAULTS["dir_path"] or appdirs.user_config_dir(utils.NAME)
    return pathlib.Path(str_path)


def load_config(dir_path=None):
    if dir_path is None:
        config_dir = get_config_dir()
    else:
        config_dir = pathlib.Path(dir_path)

    config_path = config_dir / "config.toml"

    try:
        with open(config_path) as f:
            return toml.load(f)
    except OSError:

        return {}



import csv
import subprocess
import sys


def make_reader():
    field_names = None

    def read_line(line):
        nonlocal field_names
        print("\t\t\t\t\t", line)
        reader = csv.reader([line])
        row = next(reader)

        if field_names is None:
            field_names = row
            return None

        return dict(zip(field_names, row))

    return read_line


text = """\
name,age
alice,21
bob,22
"""


def test_read_line():
    read = make_reader()

    lines = text.splitlines()

    rows = list(map(read, lines))
    # rows = list(map(dict, csv.DictReader(lines)))
    assert rows == [None, {"name": "alice", "age": "21"}, {"name": "bob", "age": "22"}]


exec_before = """\

def make_reader():
    import csv

    field_names = None

    def read_line(line):
        nonlocal field_names
        # print('\t\t\t\t\t', line)
        reader = csv.reader([line])
        row = next(reader)

        if field_names is None:
            field_names = row
            return None

        return dict(zip(field_names, row))

    return read_line

read_csv_row = make_reader()

"""


def test_py_read():

    for _ in range(10):
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "mario",
                "--exec-before",
                exec_before,
                "map",
                "read_csv_row",
            ],
            input=text.encode(),
        ).decode()

        expected = (
            "None\n{'name': 'alice', 'age': '21'}\n{'name': 'bob', 'age': '22'}\n"
        )
        assert output == expected


def test_apply_csv_dictreader_read_csv():

    for _ in range(10):
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "mario",
                "apply",
                "csv.DictReader ! [dict(y) for y in x]",
            ],
            input=text.encode(),
        ).decode()

        expected = "[{'name': 'alice', 'age': '21'}, {'name': 'bob', 'age': '22'}]\n"

        assert output == expected



import os
import sys

import click

import mario
import mario.doc

from . import app
from . import cli_tools
from . import config
from . import utils


config.DEFAULTS.update(
    config.load_config(
        dir_path=os.environ.get(f"{utils.NAME}_CONFIG_DIR".upper(), None)
    )
)


class SectionedFormatter(click.formatting.HelpFormatter):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(*args, **kwargs)

    # pylint: disable=arguments-differ
    def write_dl(self, rows, *args, **kwargs):

        if len(rows[0]) == 2:
            super().write_dl(rows)
            return

        section_name_to_commands = {}
        # pylint: disable=redefined-builtin
        for _formatted_name, subcommand, help in rows:

            section_name_to_commands.setdefault(
                getattr(subcommand, "section", "Custom"), []
            ).append((subcommand, help))

        section_name_to_spec = {s.name: s for s in self.sections}

        for section_name in sorted(
            section_name_to_commands.keys(),
            key=lambda name: section_name_to_spec.get(
                name, mario.doc.NULL_SECTION
            ).priority,
        ):
            commands = section_name_to_commands[section_name]
            section_rows = [(pair[0].name, pair[1]) for pair in commands]
            if section_name == mario.doc.UNSECTIONED:
                formatted_name = "More"
            elif section_name is None:
                formatted_name = mario.doc.NULL_SECTION.name
            else:
                formatted_name = section_name

            with super().section(formatted_name):
                super().write_dl(section_rows, *args, **kwargs)


class SectionedContext(click.Context):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(*args, **kwargs)

    def make_formatter(self):
        """Creates the formatter for the help and usage output."""
        return SectionedFormatter(
            sections=self.sections,
            width=self.terminal_width,
            max_width=self.max_content_width,
        )


class SectionedGroup(click.Group):
    def __init__(self, *args, sections, **kwargs):
        self.sections = sections
        super().__init__(self, *args, **kwargs)

    def make_context(self, info_name, args, parent=None, **extra):
        """This function when given an info name and arguments will kick
        off the parsing and create a new :class:`Context`.  It does not
        invoke the actual command callback though.

        :param info_name: the info name for this invokation.  Generally this
                          is the most descriptive name for the script or
                          command.  For the toplevel script it's usually
                          the name of the script, for commands below it it's
                          the name of the script.
        :param args: the arguments to parse as list of strings.
        :param parent: the parent context if available.
        :param extra: extra keyword arguments forwarded to the context
                      constructor.
        """
        # pylint: disable=protected-access
        for key, value in click._compat.iteritems(self.context_settings):
            if key not in extra:
                extra[key] = value
        ctx = SectionedContext(
            self, info_name=info_name, parent=parent, sections=self.sections, **extra
        )
        with ctx.scope(cleanup=False):
            self.parse_args(ctx, args)
        return ctx

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []

        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for formatted_name, subcommand in commands:

                # pylint: disable=redefined-builtin
                help = subcommand.get_short_help_str(limit)
                rows.append((formatted_name, subcommand, help))

            if rows:
                formatter.write_dl(rows)


CONTEXT_SETTINGS = {"default_map": config.DEFAULTS}

doc = f"""\
Mario: Python pipelines for your shell.

\b
Docs: https://python-mario.readthedocs.org
Addons: https://mario-addons.readthedocs.org


\b
Configuration:
  Declarative config: {config.get_config_dir() / 'config.toml'}
  Python modules: {config.get_config_dir() / 'm/*.py'}

"""


ALIASES = app.global_registry.commands


def cli_main(pairs, **kwargs):
    app.main(pairs, **kwargs)


def version_option(ctx, param, value):  # pylint: disable=unused-argument
    if not value:
        return
    click.echo("mario, version " + mario.__version__)
    sys.exit()


def build_stages(command):
    def run(ctx, **cli_params):
        out = []

        for stage in command.stages:

            mapped_stage_params = {
                remap.old.lstrip("-"): cli_params[remap.new.lstrip("-")]
                for remap in stage.remap_params
            }
            mapped_stage_params.update(stage.params)
            inject_namespace = {
                k: v for k, v in cli_params.items() if k in command.inject_values
            }
            cmd = cli.get_command(ctx, stage.command)
            out.extend(
                ctx.invoke(cmd, **mapped_stage_params, inject_values=inject_namespace)
            )
        return out

    params = command.arguments + command.options

    return cli_tools.DocumentedCommand(
        name=command.name,
        params=params,
        callback=click.pass_context(run),
        short_help=command.short_help,
        help=command.help,
        section=getattr(command, "section", None),
        hidden=command.hidden,
    )


# pylint: disable=unsupported-assignment-operation
COMMANDS = app.global_registry.cli_functions

# pylint: disable=no-member
for k, v in ALIASES.items():

    COMMANDS[k] = build_stages(v)


cli = SectionedGroup(
    result_callback=cli_main,
    chain=True,
    context_settings=CONTEXT_SETTINGS,
    params=[
        click.Option(
            ["--max-concurrent"], type=int, default=config.DEFAULTS["max_concurrent"]
        ),
        click.Option(
            ["--exec-before"],
            help="Python source code to be executed before any stage.",
            default=config.DEFAULTS["exec_before"],
        ),
        click.Option(
            ["--base-exec-before"],
            help="Python source code to be executed before any stage; typically set in the user config file. Combined with --exec-before value. ",
            default=config.DEFAULTS["base_exec_before"],
        ),
        click.Option(
            ["--version"],
            callback=version_option,
            is_flag=True,
            help="Show the version and exit.",
        ),
    ],
    help=doc,
    commands=COMMANDS,
    sections=mario.doc.SECTION_SPECS,
)



import functools


async def async_drop_falsy(items):
    async for x in items:
        if x:
            yield x


def drop_falsy(items):
    for x in items:
        if x:
            yield x


def my_max(items):
    print("hello")

    it = iter(items)

    try:
        current_max = next(it)
    except StopIteration as e:
        raise ValueError("Iterable was empty") from e
    for x in it:

        if x > current_max:
            current_max = x
    # pylint: disable=undefined-loop-variable
    return x


# def gen_max(items):
#     it = iter(items)
#     try:
#         current_max = next(it)
#     except StopIteration as e:
#         raise ValueError("Iterable was empty") from e
#     for x in it:

#         if x > current_max:
#             current_max = x
#     yield x


def gen_max(items):
    yield from [my_max(items)]


def wrap_sync_fold(function):
    @functools.wraps(function)
    def wrap(items):
        yield from [function(items)]

    return wrap


wrapped_max = wrap_sync_fold(my_max)

wrapped_drop_falsy = wrap_sync_fold(drop_falsy)



# pylint: disable=unused-argument

import subprocess
import sys
import textwrap
import time

import pytest
import requests

from . import helpers


@pytest.fixture(name="reactor")
def _reactor():
    from twisted.internet import reactor

    return reactor


@pytest.fixture(name="server")
def _server():
    # pylint: disable=fixme
    # TODO Replace subprocess with reactor
    command = [sys.executable, "-m", "tests.server"]
    proc = subprocess.Popen(command, stderr=subprocess.PIPE)
    time.sleep(1)
    yield
    proc.terminate()


def test_server(server):
    assert requests.get("http://localhost:8080/?delay=1")


def test_cli_async_map_then_apply(runner, reactor, server):
    base_url = "http://localhost:8080/?delay={}\n"

    in_stream = "".join(base_url.format(i) for i in [1, 2, 3, 4, 5] * 9)

    args = [
        "-m",
        "mario",
        "--max-concurrent",
        "100",
        "async-map",
        "await asks.get(x) ! x.json()",
        "filter",
        'x["id"] % 6 == 0',
        "map",
        "x['id']",
        "apply",
        "max(x)",
    ]

    expected = "42\n"
    limit_seconds = 6.0
    with helpers.Timer() as t:
        output = subprocess.check_output(
            [sys.executable, *args], input=in_stream.encode(), timeout=limit_seconds
        ).decode()

    assert output == expected
    assert t.elapsed < limit_seconds


def test_cli_async_map(runner, reactor, server, capsys):
    base_url = "http://localhost:8080/?delay={}\n"

    in_stream = "".join(base_url.format(i) for i in [1, 1, 5, 1])

    args = [
        "--exec-before",
        "import datetime; now=datetime.datetime.now; START_TIME=now()",
        "async-map",
        'await asks.get !  f"{types.SimpleNamespace(**x.json()).delay}"',
    ]

    expected = "1\n1\n5\n1\n"

    with helpers.Timer() as t:
        output = helpers.run(args, input=in_stream.encode()).decode()

    assert output == expected
    limit_seconds = 6.0
    assert t.elapsed < limit_seconds


def test_cli_async_map_unordered(runner, reactor, server, capsys):
    base_url = "http://localhost:8080/?delay={}\n"

    in_stream = "".join(base_url.format(i) for i in [5, 2, 3, 1, 4])

    args = [
        "async-map-unordered",
        'await asks.get !  f"{types.SimpleNamespace(**x.json()).delay}"',
    ]

    expected = "1\n2\n3\n4\n5\n"

    with helpers.Timer() as t:
        output = helpers.run(args, input=in_stream.encode()).decode()

    assert output == expected
    limit_seconds = 7.0
    assert t.elapsed < limit_seconds


def test_cli_async_reduce_fails(runner, reactor, server, capsys):
    """``reduce`` takes the name of a function as its argument, and fails otherwise."""
    base_url = "http://localhost:8080/?delay={}\n"

    in_stream = "".join(base_url.format(i) for i in [6, 2, 1])

    args = ["map", "json.loads", "reduce", "toolz.curry(operator.truediv)(*x)"]

    with pytest.raises(subprocess.CalledProcessError):
        helpers.run(args, input=in_stream.encode()).decode()


def test_cli_async_reduce_without_curry(runner, reactor, server, capsys):
    base_url = "http://localhost:8080/?delay={}\n"

    in_stream = "".join(base_url.format(i) for i in [6, 2, 1])

    args = [
        "async-map",
        'await asks.get !  f"{types.SimpleNamespace(**x.json()).delay}"',
        "map",
        "json.loads",
        "reduce",
        "operator.truediv",
    ]

    expected = "3.0\n"

    with helpers.Timer() as t:
        output = helpers.run(args, input=in_stream.encode()).decode()

    assert output == expected
    limit_seconds = 7.0
    assert t.elapsed < limit_seconds


def test_async_filter():
    """Test the async-filter command.

    There should be a better test case for this.
    """
    args = ["async-filter", '(await asks.get(x)).json()["url"].endswith(("1", "3"))']
    stdin = textwrap.dedent(
        """\
    http://httpbin.org/delay/5
    http://httpbin.org/delay/1
    http://httpbin.org/delay/2
    http://httpbin.org/delay/3
    http://httpbin.org/delay/4
    """
    )
    result = helpers.run(args, input=stdin.encode()).decode()
    expected = textwrap.dedent(
        """\
    http://httpbin.org/delay/1
    http://httpbin.org/delay/3
    """
    )
    assert result == expected



from __future__ import annotations

import json
import tempfile
import traceback
import typing as t

import attr
import docutils.nodes
import docutils.parsers.rst
import docutils.parsers.rst.directives
import marshmallow
import marshmallow.fields
import marshmallow_jsonschema
import typing_extensions as te


T = t.TypeVar("T")


FIELD_MAPPING = {
    marshmallow.fields.Bool: bool,
    marshmallow.fields.Str: str,
    marshmallow.fields.Float: float,
    marshmallow.fields.Int: int,
    marshmallow.fields.List: list,
    marshmallow.fields.Mapping: dict,
    marshmallow.fields.Dict: dict,
    marshmallow.fields.Nested: object,
}


@attr.dataclass(frozen=True, slots=True)
class Table:
    title: str
    header: t.List[str]
    body: t.List[t.List[str]] = attr.ib(factory=list)
    widths: t.Union[te.Literal["auto"], t.List[int]] = "auto"


@attr.dataclass(frozen=True, slots=True)
class Field:
    name: str
    type: t.Any  # ?
    required: bool
    default: t.Any


@attr.dataclass(frozen=True, slots=True)
class SchemaSpec:
    name: str
    fields: t.List[Field]


def quote(s):
    return '"' + s + '"'


class Marshmallow3JSONSchema(marshmallow_jsonschema.JSONSchema):
    # This class fixes incompatibilities between the parent class and Marshmallow 3.
    # It also adds the `description` field.

    # pylint: disable=unused-argument
    # pylint: disable=arguments-differ
    def wrap(self, *args, many, **kwargs):
        return super().wrap(*args, **kwargs)

    def _from_python_type(self, obj, field, pytype):
        """Get schema definition from python type."""
        json_schema = {"title": field.attribute or field.data_key or field.name}

        for key, val in marshmallow_jsonschema.base.TYPE_MAP[pytype].items():
            json_schema[key] = val

        if "description" in field.metadata:
            json_schema["description"] = field.metadata["description"]

        if field.dump_only:
            json_schema["readonly"] = True

        if field.default is not marshmallow.missing:
            json_schema["default"] = field.default

        # NOTE: doubled up to maintain backwards compatibility
        metadata = field.metadata.get("metadata", {})
        metadata.update(field.metadata)

        for md_key, md_val in metadata.items():
            if md_key == "metadata":
                continue
            json_schema[md_key] = md_val

        if isinstance(field, marshmallow.fields.List):
            json_schema["items"] = self._get_schema_for_field(obj, field.inner)
        return json_schema

    def get_properties(self, obj):
        """Fill out properties field."""
        properties = {}

        for _field_name, field in sorted(obj.fields.items()):
            schema = self._get_schema_for_field(obj, field)
            properties[field.data_key or field.name] = schema

        return properties

    def get_required(self, obj):
        """Fill out required field."""
        required = []

        for _field_name, field in sorted(obj.fields.items()):
            if field.required:
                required.append(field.data_key or field.name)

        return required or marshmallow.missing

    definition = marshmallow.fields.Method("get_definition")

    def get_definition(self, obj):
        return {obj.__doc__: []}


class SchemaDirective(docutils.parsers.rst.Directive):

    has_content = False
    required_arguments = 1
    option_spec = {
        "prog": docutils.parsers.rst.directives.unchanged_required,
        "show-nested": docutils.parsers.rst.directives.flag,
        "commands": docutils.parsers.rst.directives.unchanged,
    }

    def _get_schema(self, module_path):

        # __import__ will fail on unicode,
        # so we ensure module path is a string here.
        module_path = str(module_path)

        try:
            module_name, attr_name = module_path.split(":", 1)
        except ValueError:  # noqa
            raise self.error(
                '"{}" is not of format "module:schema"'.format(module_path)
            )

        try:
            mod = __import__(module_name, globals(), locals(), [attr_name])
        except (Exception, SystemExit) as exc:  # noqa
            err_msg = 'Failed to import "{}" from "{}". '.format(attr_name, module_name)
            if isinstance(exc, SystemExit):
                err_msg += "The module appeared to call sys.exit()"
            else:
                err_msg += "The following exception was raised:\n{}".format(
                    traceback.format_exc()
                )

            raise self.error(err_msg)

        if not hasattr(mod, attr_name):
            raise self.error(
                'Module "{}" has no attribute "{}"'.format(module_name, attr_name)
            )

        schema = getattr(mod, attr_name)

        if not issubclass(schema, marshmallow.Schema):
            raise self.error(
                '"{}" of type "{}" is not derived from '
                '"marshmallow.Schema"'.format(schema, module_path)
            )
        return schema

    def _get_inner(self, field):
        if hasattr(field, "inner"):
            return self._get_inner(field.inner)

        if hasattr(field, "nested"):
            return self._get_inner(field.nested)

        return field

    def _build_section(self, schema):

        mm_json = Marshmallow3JSONSchema().dump(schema)

        # Title

        name = type(schema).__name__

        section = docutils.nodes.section(
            "",
            docutils.nodes.title(text=name),
            ids=[docutils.nodes.make_id(name)],
            names=[docutils.nodes.fully_normalize_name(name)],
        )

        # Summary

        source_name = type(schema).__name__
        result = docutils.statemachine.ViewList()

        file = tempfile.NamedTemporaryFile(mode="wt", delete=False, prefix="sphinx")

        json.dump(mm_json, file, indent=4)
        file.close()

        lines = []
        lines += [f".. jsonschema:: {file.name}\n"]

        for line in lines:
            result.append(line, source_name)

        self.state.nested_parse(result, 0, section)

        subsections = []
        for field_name in schema.fields:
            field = schema.declared_fields[field_name]
            inner = self._get_inner(field)

            if isinstance(inner, type) and issubclass(inner, marshmallow.Schema):
                subsections += self._build_section(inner())

        return [section] + subsections

    def run(self):
        # pylint: disable=attribute-defined-outside-init
        self.env = self.state.document.settings.env

        schema = self._get_schema(self.arguments[0])

        return self._build_section(schema())


def setup(app):
    app.add_directive("marshmallow", SchemaDirective)



"""Functions for write commands."""

from __future__ import annotations

import csv
import io
import itertools
import typing as t

import yaml


def write_csv_dicts(rows: t.Iterable[t.Dict], header: bool, dialect: str) -> str:
    """Write iterable of dicts to csv."""
    file = io.StringIO()

    iterator = iter(rows)
    first_row = next(iterator)
    fieldnames = list(first_row.keys())
    writer = csv.DictWriter(file, fieldnames=fieldnames, dialect=dialect)
    if header:
        writer.writeheader()
    writer.writerows(itertools.chain([first_row], iterator))

    return file.getvalue()


def write_csv_tuples(rows: t.Iterable[t.Tuple], dialect: str) -> str:
    """Write iterable of tuples to csv."""
    file = io.StringIO()

    writer = csv.writer(file, dialect)
    writer.writerows(rows)

    return file.getvalue()


def write_yaml(data) -> str:
    """Write data to yaml string."""
    file = io.StringIO()
    yaml.dump(data, file)
    return file.getvalue()



from tests import helpers


def test_eval_1_is_fast():
    """``eval `` should be very quick."""
    # This should ideally be below 0.1s.
    with helpers.Timer(max=0.6):
        output = helpers.run(["eval", "1"]).decode()

    assert output == "1\n"



# pylint: disable=protected-access
# pylint: disable=unused-argument

from __future__ import generator_stop

import os
import subprocess
import sys
import textwrap

import hypothesis
import pytest
from tests import helpers

import mario
import mario.app
import mario.cli
from mario import utils


hypothesis.settings.register_profile("ci", max_examples=1000)
hypothesis.settings.register_profile("dev", max_examples=10)
hypothesis.settings.register_profile(
    "debug", max_examples=10, verbosity=hypothesis.Verbosity.verbose
)
hypothesis.settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))


def assert_exception_equal(e1, e2):
    # pylint: disable=unidiomatic-typecheck
    assert type(e1) == type(e2)
    assert e1.args == e2.args


@pytest.mark.parametrize("option", ["--invented-option", "-J"])
def test_raises_on_nonexistent_option(option, runner):
    args = [option, "print"]
    in_stream = "a.b.c\n"

    result = runner.invoke(mario.cli.cli, args, input=in_stream)
    assert result.exception


def test_eval_main(capsys):
    mario.app.main([[{"name": "eval", "code": "1+1", "parameters": {}}]])
    assert capsys.readouterr().out == "2\n"


def test_eval_cli():
    assert helpers.run(["eval", "1+1"]).decode() == "2\n"


def test_chain():
    expected = "[1, 2]\n"
    result = helpers.run(["eval", "[[1, 2]]", "chain"]).decode()
    assert result == expected, (result, expected)


def test_cli_version(runner):
    args = ["--version"]
    result = helpers.run(args).decode()
    assert result == f"mario, version {mario.__version__}\n"


def test_config_file(tmp_path):
    config_body = """
    exec_before = "from collections import Counter as C"
    """

    config_file_path = tmp_path / "config.toml"

    config_file_path.write_text(config_body)

    args = ["apply", "C(x)"]
    stdin = "1\n2\n".encode()
    env = dict(os.environ)
    env.update({f"{utils.NAME}_CONFIG_DIR".upper().encode(): str(tmp_path).encode()})
    output = helpers.run(args, input=stdin, env=env).decode()
    assert output.startswith("Counter")


def test_exec_before():
    exec_before = textwrap.dedent(
        """\
    import csv
    def func(line):
        return next(csv.reader([line]))
    """
    )

    assert (
        helpers.run(
            ["--exec-before", exec_before, "map", "func"], input=b"a,b\n"
        ).decode()
        == "['a', 'b']\n"
    )


def test_stage_exec_before():
    assert helpers.run(["eval", "--exec-before", "a=1", "a"]).decode() == "1\n"


def test_meta_pip_command():
    """``meta pip`` command accepts pip subcommands and options."""
    result = helpers.run(["meta", "pip", "install", "--help"]).decode()
    expected = (
        "pip install [options] <requirement specifier> [package-index-options] ..."
    )
    assert expected in result


def test_meta_test_command_fail(tmp_path, tmp_env):
    """Get a failing result when a test fails."""

    text = r"""
    [[command]]
    name = "mario-failing-test"
    short_help = "An internal test command for mario"
    help = "An internal test command for mario"

    [[command.stages]]
    command = "eval"
    params = {code="1"}

    [[command.tests]]
    invocation = ["mario-failing-test"]
    input = ""
    output = "2\n"

    """
    (tmp_path / "config.toml").write_text(text)

    proc = subprocess.run(
        [sys.executable, "-m", "python", "meta", "test", "-vvvvv", "-pno:sugar"],
        env=tmp_env,
        check=False,
        capture_output=True,
    )
    assert proc.returncode != 0


def test_meta_test_command_pass(tmp_path, tmp_env):
    """Get a failing result when a test fails."""

    text = r"""
    [[command]]
    name = "mario-passing-test"
    short_help = "An internal test command for mario"
    help = "An internal test command for mario"

    [[command.stages]]
    command = "eval"
    params = {code="1"}

    [[command.tests]]
    invocation = ["mario-passing-test"]
    input = ""
    output = "1\n"
    """
    (tmp_path / "config.toml").write_text(text)

    proc = subprocess.run(
        [
            sys.executable,
            "-m",
            "mario",
            "meta",
            "test",
            "-vvvv",
            "-pno:sugar",
            "--tb=long",
        ],
        env=tmp_env,
        capture_output=True,
    )
    assert proc.returncode == 0



import textwrap

from tests import helpers

import mario.doc


def test_rst2text():
    source = textwrap.dedent(
        """\
    ============
    Introduction
    ============

    Hello world.

    .. code-block:: bash

        $ echo Greetings.


    """
    )

    result = mario.doc.rst2text(source)
    expected = textwrap.dedent(
        """\
    Introduction
    ============

    Hello world.

       $ echo Greetings.
    """
    )
    assert result == expected


def test_help():
    """CLI help is formatted as plain text not rst and includes a demo."""
    result = helpers.run(["map", "--help"]).decode()

    assert "$ mario map" in result
    assert ".. code-block" not in result


def test_render_short_help(tmp_path, tmp_env):
    """CLI renders ``short_help`` if ``help`` is missing."""
    result = helpers.run(["map", "--help"]).decode()
    text = """
    [[command]]
    name = "my-command"
    short_help = "This is the short help."

    [[command.stages]]
    command = "eval"
    params = {code='1'}
    """

    (tmp_path / "config.toml").write_text(text)

    result = helpers.run(["my-command", "--help"], env=tmp_env).decode()

    assert "This is the short help." in result



import pathlib
import re

import setuptools


PROJECT_ROOT = pathlib.Path(__file__).parent

with open(PROJECT_ROOT / "requirements.in") as f:
    INSTALL_REQUIRES = f.read().splitlines()

with open(PROJECT_ROOT / "dev-requirements.txt") as f:
    DEV_INSTALL_REQUIRES = f.read().splitlines()

setuptools.setup(
    name="mario",
    version="0.0.152",
    description="Shell pipes for Python.",
    long_description=open(PROJECT_ROOT / "README.rst").read(),
    long_description_content_type="text/x-rst",
    author="mario contributors",
    author_email="mario@example.com",
    packages=setuptools.find_packages("src"),
    package_dir={"": "src"},
    url="https://github.com/python-mario/mario",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    python_requires=">=3.7",
    install_requires=INSTALL_REQUIRES,
    entry_points={
        "console_scripts": ["mario = mario.cli:cli"],
        "mario_plugins": ["basic = mario.plugins"],
    },
    extras_require={"dev": DEV_INSTALL_REQUIRES},
)



#!/usr/bin/env python
"""Command line pipes in python."""

from __future__ import generator_stop

import itertools
import typing

import trio


T = typing.TypeVar("T")
U = typing.TypeVar("U")

_PYPE_VALUE = "_PYPE_VALUE"

BUFSIZE = 2 ** 14
counter = itertools.count()
_RECEIVE_SIZE = 4096  # pretty arbitrary


class TerminatedFrameReceiver:
    """Parse frames out of a Trio stream, where each frame is terminated by a
    fixed byte sequence.

    For example, you can parse newline-terminated lines by setting the
    terminator to b"\n".

    This uses some tricks to protect against denial of service attacks:

    - It puts a limit on the maximum frame size, to avoid memory overflow; you
    might want to adjust the limit for your situation.

    - It uses some algorithmic trickiness to avoid "slow loris" attacks. All
      algorithms are amortized O(n) in the length of the input.

    """

    def __init__(
        self,
        stream: trio.abc.ReceiveStream,
        terminator: bytes,
        max_frame_length: int = 16384,
    ) -> None:
        self.stream = stream
        self.terminator = terminator
        self.max_frame_length = max_frame_length
        self._buf = bytearray()
        self._next_find_idx = 0

    async def receive(self) -> bytearray:
        while True:
            terminator_idx = self._buf.find(self.terminator, self._next_find_idx)
            if terminator_idx < 0:
                # no terminator found
                if len(self._buf) > self.max_frame_length:
                    raise ValueError("frame too long")
                # next time, start the search where this one left off
                self._next_find_idx = max(0, len(self._buf) - len(self.terminator) + 1)
                # add some more data, then loop around
                more_data = await self.stream.receive_some(_RECEIVE_SIZE)
                if more_data == b"":
                    if self._buf:
                        raise ValueError("incomplete frame")
                    raise trio.EndOfChannel
                self._buf += more_data
            else:
                # terminator found in buf, so extract the frame
                frame = self._buf[:terminator_idx]
                # Update the buffer in place, to take advantage of bytearray's
                # optimized delete-from-beginning feature.
                del self._buf[: terminator_idx + len(self.terminator)]
                # next time, start the search from the beginning
                self._next_find_idx = 0
                return frame

    def __aiter__(self) -> "TerminatedFrameReceiver":
        return self

    async def __anext__(self) -> bytearray:
        try:
            return await self.receive()
        except trio.EndOfChannel:
            raise StopAsyncIteration



#!/usr/bin/env python
"""Command line pipes in python."""

from __future__ import generator_stop

import collections
import functools
import os
from typing import AsyncIterable
from typing import List

import async_exit_stack
import attr
import trio

from . import asynch
from . import config
from . import interfaces
from . import interpret
from . import plug


async def call_traversal(
    context,  # pylint: disable=unused-argument
    traversal: interfaces.Traversal,
    items: AsyncIterable,
    exit_stack: async_exit_stack.AsyncExitStack,
):
    runtime_parameters = {"items": items, "exit_stack": exit_stack}

    calculated_params = traversal.plugin_object.calculate_more_params(traversal)

    available_params = collections.ChainMap(
        calculated_params,
        runtime_parameters,
        traversal.specific_invocation_params,
        traversal.global_invocation_options.global_options,
    )

    args = {
        param: available_params[param]
        for param in traversal.plugin_object.required_parameters
    }
    return await traversal.plugin_object.traversal_function(**args)


async def program_runner(
    traversals: List[interfaces.Traversal],
    items: AsyncIterable,
    context: interfaces.Context,
):
    async with async_exit_stack.AsyncExitStack() as stack:
        for traversal in traversals:
            items = await call_traversal(context, traversal, items, stack)

        return stack.pop_all(), items


async def async_main(basic_traversals, **kwargs):
    # pylint: disable=protected-access
    stream = trio._unix_pipes.PipeReceiveStream(os.dup(0))
    receiver = asynch.TerminatedFrameReceiver(stream, b"\n")
    items = (item.decode() async for item in receiver)

    # pylint: disable=no-member
    global_context = interfaces.Context(global_registry.global_options.copy())
    global_context.global_options.update(config.DEFAULTS)
    global_context.global_options.update(kwargs)

    global_context.global_options[
        "global_namespace"
    ] = interpret.build_global_namespace(
        global_context.global_options["base_exec_before"]
    )

    global_context.global_options["global_namespace"].update(
        interpret.build_global_namespace(global_context.global_options["exec_before"])
    )
    traversals = []

    for bt in basic_traversals:

        for d in bt:
            # pylint: disable=fixme
            # TODO Make classes or use pyrsistent.
            traversal_namespace = {
                **global_context.global_options["global_namespace"],
                **d["parameters"].get("inject_values", {"HELLO": "WORLD"}),
            }
            traversal_context = attr.evolve(
                global_context,
                global_options=dict(
                    global_context.global_options, global_namespace=traversal_namespace
                ),
            )
            # pylint: disable=unsubscriptable-object
            traversal = interfaces.Traversal(
                global_invocation_options=traversal_context,
                specific_invocation_params=d,
                plugin_object=global_registry.traversals[d["name"]],
            )
            traversals.append(traversal)

    stack, items = await program_runner(traversals, items, global_context)

    async with stack:
        async for item in items:
            print(item)


def main(pairs, **kwargs):
    trio.run(functools.partial(async_main, pairs, **kwargs))


global_registry = plug.make_global_registry()



import subprocess
import sys

import click

from mario import cli_tools
from mario import doc
from mario import interpret
from mario import plug
from mario import traversals


registry = plug.Registry()


def calculate_function(traversal, howcall=None):
    if howcall is None:
        howcall = traversal.specific_invocation_params.get("howcall")
    if howcall is None:
        howcall = interpret.HowCall.SINGLE

    global_namespace = traversal.global_invocation_options.global_options[
        "global_namespace"
    ].copy()

    if "exec_before" in traversal.specific_invocation_params["parameters"]:
        global_namespace.update(
            interpret.build_global_namespace(
                traversal.specific_invocation_params["parameters"]["exec_before"]
            )
        )

    if "code" in traversal.specific_invocation_params:

        return {
            "function": interpret.build_function(
                traversal.specific_invocation_params["code"],
                global_namespace=global_namespace,
                howcall=howcall,
            )
        }

    return {"function": None}


def calculate_reduce(traversal):

    function = interpret.build_function(
        traversal.specific_invocation_params["code"],
        traversal.global_invocation_options.global_options["global_namespace"],
        howcall=interpret.HowCall.VARARGS,
    )

    return {"function": function}


@registry.add_traversal("map", calculate_more_params=calculate_function)
async def map(
    function, items, exit_stack, max_concurrent
):  # pylint: disable=redefined-builtin
    """
    Run code on each input item.

    Each item is handled in the order it was received, and also output in the
    same order. For less strict ordering and asynchronous execution, see
    ``async-map`` and ``async-map-unordered``.

    For example,

    .. code-block:: bash

        $ mario map 'x*2' <<EOF
        a
        b
        c
        EOF
        aa
        bb
        cc

    """
    return await exit_stack.enter_async_context(
        traversals.sync_map(function, items, max_concurrent)
    )


@registry.add_traversal("async_map", calculate_more_params=calculate_function)
async def async_map(function, items, exit_stack, max_concurrent):
    """
    Run code on each input item asynchronously.

    The order of inputs is retained in the outputs. However, the order of inputs
    does not determine the order in which each input is handled, only the order
    in which its result is emitted. To keep the order in which each input is
    handled, use the synchronous version, ``map``.

    In this example, we make requests that have a server-side delay of specified
    length. The input order is retained in the output by holding each item until
    its precedents are ready.

    .. code-block:: bash


           $ mario async-map 'await asks.get ! x.json()["url"]'  <<EOF
           http://httpbin.org/delay/5
           http://httpbin.org/delay/1
           http://httpbin.org/delay/2
           http://httpbin.org/delay/3
           http://httpbin.org/delay/4
           EOF
           https://httpbin.org/delay/5
           https://httpbin.org/delay/1
           https://httpbin.org/delay/2
           https://httpbin.org/delay/3
           https://httpbin.org/delay/4

    """
    return await exit_stack.enter_async_context(
        traversals.async_map(function, items, max_concurrent)
    )


@registry.add_traversal("async_map_unordered", calculate_more_params=calculate_function)
async def async_map_unordered(function, items, exit_stack, max_concurrent):
    """
    Run code on each input item asynchronously, without retaining input order.

    Each result is emitted in the order it becomes ready, regardless of input
    order. Input order is also ignored when determining in which order to
    *start* handling each item. Results start emitting as soon as the first one
    is ready. It also saves memory because it doesn't require accumulating
    results while waiting for previous items to become ready. For stricter
    ordering, see ``map`` or ``async_map``.

    In this example, we make requests that have a server-side delay of specified
    length. The input order is lost but the results appear immediately as they
    are ready (the delay length determines the output order):

    .. code-block:: bash

           $ mario async-map-unordered 'await asks.get ! x.json()["url"]'  <<EOF
           http://httpbin.org/delay/5
           http://httpbin.org/delay/1
           http://httpbin.org/delay/2
           http://httpbin.org/delay/3
           http://httpbin.org/delay/4
           EOF
           https://httpbin.org/delay/1
           https://httpbin.org/delay/2
           https://httpbin.org/delay/3
           https://httpbin.org/delay/4
           https://httpbin.org/delay/5

    """
    return await exit_stack.enter_async_context(
        traversals.async_map_unordered(function, items, max_concurrent)
    )


@registry.add_traversal("filter", calculate_more_params=calculate_function)
async def filter(
    function, items, exit_stack, max_concurrent
):  # pylint: disable=redefined-builtin
    """
    Keep input items that satisfy a condition.

    Order of input items is retained in the output.

    For example,

    .. code-block:: bash

        $ mario filter 'x > "c"' <<EOF
        a
        b
        c
        d
        e
        f
        EOF
        d
        e
        f
    """

    return await exit_stack.enter_async_context(
        traversals.sync_filter(function, items, max_concurrent)
    )


@registry.add_traversal("async_filter", calculate_more_params=calculate_function)
async def async_filter(function, items, exit_stack, max_concurrent):
    """
    Keep input items that satisfy an asynchronous condition.

    For example,

    .. code-block:: bash

        $ mario async-filter '(await asks.get(x)).json()["url"].endswith(("1", "3"))'  <<EOF
        http://httpbin.org/delay/5
        http://httpbin.org/delay/1
        http://httpbin.org/delay/2
        http://httpbin.org/delay/3
        http://httpbin.org/delay/4
        EOF
        http://httpbin.org/delay/1
        http://httpbin.org/delay/3

    """
    return await exit_stack.enter_async_context(
        traversals.async_filter(function, items, max_concurrent)
    )


@registry.add_traversal("apply", calculate_more_params=calculate_function)
async def apply(function, items):
    """
    Apply code to the iterable of items.

    The code should take an iterable and it will be called with the input items.
    The items iterable will be converted to a list before the code is called, so
    it doesn't work well on very large streams.

    For example,

    .. code-block:: bash

        $ mario map int apply sum <<EOF
        10
        20
        30
        EOF
        60

    """
    return traversals.AsyncIterableWrapper([await function([x async for x in items])])


@registry.add_traversal("async_apply", calculate_more_params=calculate_function)
async def async_apply(function, items):
    """Apply code to an async iterable of items.

    The code should take an async iterable.
    """
    return await traversals.async_apply(function, items)


# pylint: disable=redefined-builtin
@registry.add_traversal(
    "eval",
    calculate_more_params=lambda x: calculate_function(
        x, howcall=interpret.HowCall.NONE
    ),
)
async def eval(function):
    """
    Evaluate a Python expression.

    No input items are used.

    For example,

    .. code-block:: bash

        $ mario eval 1+1
        2
    """
    return traversals.AsyncIterableWrapper([await function(None)])


@registry.add_traversal("reduce", calculate_more_params=calculate_reduce)
async def reduce(function, items, exit_stack, max_concurrent):
    """
    Reduce input items with code that takes two arguments, similar to ``functools.reduce``.

    For example,

    .. code-block:: bash

        $ mario reduce map int operator.mul <<EOF
        1
        2
        3
        4
        5
        EOF
        120

    """
    return await exit_stack.enter_async_context(
        traversals.async_reduce(function, items, max_concurrent)
    )


@registry.add_traversal(
    "chain",
    calculate_more_params=lambda x: calculate_function(
        x, howcall=interpret.HowCall.NONE
    ),
)
async def chain(items, exit_stack):
    """
    Concatenate a sequence of input iterables together into one long iterable.

    Converts an iterable of iterables of items into an iterable of items, like `itertools.chain.from_iterable <https://docs.python.org/3/library/itertools.html#itertools.chain.from_iterable>`_.

    For example,

    .. code-block:: bash

        $ mario eval '[[1,2]]'
        [[1, 2]]


        $ mario eval '[[1, 2]]' chain
        [1, 2]

    """
    return await exit_stack.enter_async_context(traversals.sync_chain(items))


@registry.add_traversal(
    "async-chain",
    calculate_more_params=lambda x: calculate_function(
        x, howcall=interpret.HowCall.NONE
    ),
)
async def async_chain(items, exit_stack):
    """
    Concatenate a sequence of input async iterables into one long async iterable.

    Converts an async iterable of async iterables of items into an async
    iterable of items, like `itertools.chain.from_iterable <https://docs.python.org/3/library/itertools.html#itertools.chain.from_iterable>`_
    for async iterables.

    """
    return await exit_stack.enter_async_context(traversals.async_chain(items))


subcommands = [
    cli_tools.DocumentedCommand(
        "map",
        help=map.__doc__,
        short_help="Call code on each line of input.",
        section="Traversals",
    ),
    cli_tools.DocumentedCommand(
        "async-map",
        help=async_map.__doc__,
        short_help="Call code on each line of input.",
        section="Async traversals",
    ),
    cli_tools.DocumentedCommand(
        "apply",
        help=apply.__doc__,
        short_help="Call code on input as a sequence.",
        section="Traversals",
    ),
    cli_tools.DocumentedCommand(
        "async-apply",
        help=async_apply.__doc__,
        short_help="Call code asynchronously on input as a sequence.",
        section="Async traversals",
    ),
    cli_tools.DocumentedCommand(
        "filter",
        help=filter.__doc__,
        short_help="Call code on each line of input and exclude false values.",
        section="Traversals",
    ),
    cli_tools.DocumentedCommand(
        "async-filter",
        help=async_filter.__doc__,
        short_help="Async call code on each line of input and exclude false values.",
        section="Async traversals",
    ),
    cli_tools.DocumentedCommand(
        "async-map-unordered",
        help=async_map_unordered.__doc__,
        short_help="Call code on each line of input, ignoring order of input items.",
        section="Async traversals",
    ),
    cli_tools.DocumentedCommand(
        "eval",
        help=eval.__doc__,
        short_help="Evaluate a python expression code",
        section="Traversals",
    ),
]


def build_callback(sub_command):
    def callback(code, autocall, **parameters):
        if autocall:
            howcall = interpret.HowCall.SINGLE
        else:
            howcall = interpret.HowCall.NONE

        return [
            {
                "name": sub_command.name.replace("-", "_"),
                "howcall": howcall,
                "code": code,
                "parameters": parameters,
            }
        ]

    return callback


option_exec_before = click.option(
    "--exec-before", help="Execute code in the function's global namespace."
)

for subcommand in subcommands:

    subcommand.params = [
        click.Option(
            ["--autocall/--no-autocall"],
            is_flag=True,
            default=True,
            help='Automatically call the function if "x" does not appear in the expression. '
            "Allows ``map len`` instead of ``map len(x)``.",
        ),
        click.Argument(["code"]),
    ]
    subcommand.callback = build_callback(subcommand)
    subcommand = option_exec_before(subcommand)
    # pylint: disable=fixme
    # TODO: add_cli and add_traversal should be the non-decorator form
    registry.add_cli(name=subcommand.name)(subcommand)


@registry.add_cli(name="reduce")
@click.command(  # type: ignore
    "reduce",
    cls=cli_tools.DocumentedCommand,
    section="Traversals",
    short_help="Reduce a sequence with a function like ``operator.mul``.",
    help=reduce.__doc__,
)
@option_exec_before
@click.argument("function_name")
def _reduce(function_name, **parameters):
    return [
        {
            "code": f"toolz.curry({function_name})",
            "name": "reduce",
            "parameters": parameters,
        }
    ]


# @registry.add_cli(name="eval")
# @click.command("eval", short_help="Call <code> without any input.")
# @option_exec_before
# @click.argument("expression")
# def _eval(expression, **parameters):
#     return [{"code": expression, "name": "eval", "parameters": parameters}]


more_commands = [
    cli_tools.DocumentedCommand(
        "chain",
        callback=lambda **kw: [{"name": "chain", "parameters": kw}],
        help=chain.__doc__,
        short_help="Expand iterable of iterables of items into an iterable of items.",
        section="Traversals",
    ),
    cli_tools.DocumentedCommand(
        "async-chain",
        callback=lambda **kw: [{"name": "async-chain", "parameters": kw}],
        short_help="Expand iterable of async iterables into an iterable of items.",
        help=async_chain.__doc__,
        section="Async traversals",
    ),
]
for cmd in more_commands:
    registry.add_cli(name=cmd.name)(cmd)


meta = click.Group("meta", chain=True)
meta.section = doc.UNSECTIONED  # type: ignore
meta.sections = None  # type: ignore
meta.help = "Commands about using mario."


@meta.command(
    context_settings=dict(ignore_unknown_options=True),
    cls=cli_tools.DocumentedCommand,
    section=doc.UNSECTIONED,
)
@click.argument("pip_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def pip(ctx, pip_args):
    """
    Run ``pip`` in the environment that mario is installed into.

    Arguments are forwarded to ``pip``.
    """
    cli_args = [sys.executable, "-m", "pip"] + list(pip_args)
    ctx.exit(subprocess.run(cli_args).returncode)


@meta.command(
    "test",
    cls=cli_tools.DocumentedCommand,
    section=doc.UNSECTIONED,
    context_settings=dict(ignore_unknown_options=True),
)
@click.argument("pytest_args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def run_tests(ctx, pytest_args):
    """
    Run all declarative command tests from plugins and config.

    Executes each test in the ``command.tests`` field with pytest.

    Default pytest args: ``-vvv --tb=short``
    """

    import tempfile
    import textwrap

    pytest_args = list(pytest_args) or ["-vvv", "--tb=short"]

    source = textwrap.dedent(
        """\
    import subprocess
    import sys

    import pytest

    import mario.app

    COMMANDS = mario.app.global_registry.commands.values()  # pylint: disable=no-member
    TEST_SPECS = [test for command in COMMANDS for test in command.tests]


    @pytest.mark.parametrize(\"test_spec\", TEST_SPECS, ids=lambda ts: str(list(ts.invocation)))
    def test_command(test_spec):

        output = subprocess.check_output(
            [sys.executable, \"-m\", \"mario\"] + list(test_spec.invocation),
            input=test_spec.input.encode(),
        ).decode()
        assert output == test_spec.output

    """
    )
    f = tempfile.NamedTemporaryFile("wt", suffix=".py", delete=False)
    f.write(source)
    f.close()
    args = [sys.executable, "-m", "pytest"] + pytest_args + [f.name]
    proc = subprocess.run(args)
    ctx.exit(proc.returncode)


registry.add_cli(name="meta")(meta)



import pathlib
import subprocess
import sys
import time

import attr

import mario.exceptions


TESTS_DIR = pathlib.Path(__file__).parent


def run(args, capture_output=True, check=True, **kwargs):
    args = [sys.executable, "-m", "mario"] + args
    proc = subprocess.run(args, capture_output=capture_output, check=check, **kwargs)
    return proc.stdout


@attr.s(auto_exc=True)
class TimerMaxExceeded(mario.exceptions.MarioException):
    """Raised if the timer max is exceeded."""

    timer = attr.ib()


@attr.s
class Timer:
    """Time the body of the context manager.


    Args:
        max (float): Maximum allowed time. An exception will be raised if the
                     total duration exceeded this value. The body will not be
                     interrupted by the timer; ``max`` is only checked
                     afterwards.
        elapsed (float): Total elapsed seconds during the context manager body.
                         Set to ``None`` until the context manager exits.


    Raises:
        TimerMaxExceeded: If the maximum is exceeded when the context manager
                          exits.
    """

    max = attr.ib(default=float("inf"))
    elapsed = attr.ib(default=None)
    _start = attr.ib(default=None)
    _end = attr.ib(default=None)

    def __enter__(self):
        self._start = time.monotonic()
        return self

    def __exit__(self, *args):
        self._end = time.monotonic()
        self.elapsed = self._end - self._start

        if self.elapsed > self.max:
            raise TimerMaxExceeded(self)



from __future__ import annotations
from __future__ import generator_stop

import itertools
import types
import typing as t
from typing import AsyncIterable
from typing import AsyncIterator
from typing import Awaitable
from typing import Callable
from typing import Iterable

import async_generator
import trio


T = t.TypeVar("T")
U = t.TypeVar("U")

_PYPE_VALUE = "_PYPE_VALUE"

BUFSIZE = 2 ** 14
counter = itertools.count()
_RECEIVE_SIZE = 4096  # pretty arbitrary


async def aenumerate(items, start=0):
    i = start
    async for x in items:
        yield i, x
        i += 1


class AsyncIterableWrapper:
    def __init__(self, iterable):
        self.iterable = iter(iterable)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.iterable)
        except StopIteration:
            raise StopAsyncIteration


async def async_apply(function, data):
    return await function(data)


@async_generator.asynccontextmanager
async def async_map(
    function: Callable[[T], Awaitable[U]], iterable: AsyncIterable[T], max_concurrent
) -> AsyncIterator[AsyncIterable[U]]:
    # pylint: disable=unsubscriptable-object
    send_result, receive_result = trio.open_memory_channel[U](0)
    limiter = trio.CapacityLimiter(max_concurrent)

    async def wrapper(prev_done: trio.Event, self_done: trio.Event, item: T) -> None:

        # pylint: disable=not-async-context-manager
        async with limiter:
            result = await function(item)

        await prev_done.wait()
        await send_result.send(result)
        self_done.set()

    async def consume_input(nursery) -> None:
        prev_done = trio.Event()
        prev_done.set()
        async for item in iterable:
            self_done = trio.Event()
            nursery.start_soon(wrapper, prev_done, self_done, item)
            prev_done = self_done
        await prev_done.wait()
        await send_result.aclose()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(consume_input, nursery)
        yield receive_result
        nursery.cancel_scope.cancel()


@async_generator.asynccontextmanager
async def sync_map(
    function: Callable[[T], Awaitable[U]],
    iterable: AsyncIterable[T],
    max_concurrent,  # pylint: disable=unused-argument
) -> AsyncIterator[AsyncIterable[U]]:
    yield (await function(item) async for item in iterable)


@async_generator.asynccontextmanager
async def sync_chain(iterable: AsyncIterable[Iterable], **_kwargs):
    yield (item async for subiterable in iterable for item in subiterable)


@async_generator.asynccontextmanager
async def async_chain(iterable: AsyncIterable[AsyncIterable], **_kwargs):
    yield (item async for subiterable in iterable async for item in subiterable)


@async_generator.asynccontextmanager
async def sync_filter(
    function: Callable,
    iterable: AsyncIterable,
    max_concurrent,  # pylint: disable=unused-argument
) -> AsyncIterator[AsyncIterable[U]]:
    yield (item async for item in iterable if await function(item))


@async_generator.asynccontextmanager
async def async_map_unordered(
    function: Callable[[T], Awaitable[U]], iterable: AsyncIterable[T], max_concurrent
) -> AsyncIterator[AsyncIterable[U]]:
    # pylint: disable=unsubscriptable-object
    send_result, receive_result = trio.open_memory_channel[U](0)
    limiter = trio.CapacityLimiter(max_concurrent)
    remaining_tasks: t.Set[int] = set()

    async def wrapper(task_id: int, item: T) -> None:
        # pylint: disable=not-async-context-manager
        async with limiter:
            result = await function(item)

        await send_result.send(result)
        remaining_tasks.remove(task_id)

    async def consume_input(nursery) -> None:

        async for task_id, item in aenumerate(iterable):
            remaining_tasks.add(task_id)
            nursery.start_soon(wrapper, task_id, item)

        while remaining_tasks:
            await trio.sleep(0)

        await send_result.aclose()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(consume_input, nursery)
        yield receive_result
        nursery.cancel_scope.cancel()


@async_generator.asynccontextmanager
async def async_filter(
    function: Callable[[T], Awaitable[T]], iterable: AsyncIterable[T], max_concurrent
) -> AsyncIterator[AsyncIterable[T]]:
    # pylint: disable=unsubscriptable-object
    send_result, receive_result = trio.open_memory_channel[T](0)

    limiter = trio.CapacityLimiter(max_concurrent)

    async def wrapper(prev_done: trio.Event, self_done: trio.Event, item: T) -> None:
        # pylint: disable=not-async-context-manager
        async with limiter:
            result = await function(item)

        await prev_done.wait()
        if result:
            await send_result.send(item)
        self_done.set()

    async def consume_input(nursery) -> None:
        prev_done = trio.Event()
        prev_done.set()
        async for item in iterable:
            self_done = trio.Event()
            nursery.start_soon(wrapper, prev_done, self_done, item)
            prev_done = self_done
        await prev_done.wait()
        await send_result.aclose()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(consume_input, nursery)
        yield receive_result
        nursery.cancel_scope.cancel()


SENTINEL = object()


@async_generator.asynccontextmanager
async def async_reduce(
    function: Callable[[T, U], Awaitable[U]],
    iterable: AsyncIterable[T],
    max_concurrent,
    initializer=SENTINEL,
) -> AsyncIterator[AsyncIterable[U]]:
    # pylint: disable=unsubscriptable-object
    send_result, receive_result = trio.open_memory_channel[U](0)
    limiter = trio.CapacityLimiter(max_concurrent)

    collected_result = initializer

    async def wrapper(prev_done: trio.Event, self_done: trio.Event, item: T) -> None:
        nonlocal collected_result

        input_item = await wait_for(item)

        if collected_result is SENTINEL:
            # We are working on the first item, and initializer was not set.
            collected_result = input_item

        else:

            async with limiter:  # pylint: disable=not-async-context-manager
                collected_result = await function(collected_result, input_item)

        await prev_done.wait()
        self_done.set()

    async def consume_input(nursery) -> None:
        prev_done = trio.Event()
        prev_done.set()
        async for item in iterable:
            self_done = trio.Event()
            nursery.start_soon(wrapper, prev_done, self_done, item)
            prev_done = self_done
        await prev_done.wait()
        await send_result.send(collected_result)
        await send_result.aclose()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(consume_input, nursery)
        yield receive_result
        nursery.cancel_scope.cancel()


async def wait_for(x):
    if isinstance(x, types.CoroutineType):
        return await x
    return x
